function [min_x, min_val, bestconflist, model, incumbent_theta_idx, func, options, rundata] = smbo(func, options, logHeader)
% Function   : SMAC Version 1.1
% Written by : Frank Hutter (hutter@cs.ubc.ca)
% Created on : 11/05/2011
% Last Update: 08/06/2011
% Purpose    : Sequential model-based optimization of expensive mixed
% continuous/discrete blackbox functions
%
% This code comes with no guarantee or warranty of any kind.
%
% For details on SMAC, see our LION-2011 paper.
%
% Input parameters:
%   func: a struct describing the function to be optimized, with the
%         following fields. 
%         Assertions: - At least one of func.cont or func.cat is set and non-empty
%                     - func.cat and func.cont are disjoint (if set)
%                     - union(func.cat, func.cont) = {1:max(max(func.cont),max(func.cat))}
%                     - If func.cont is set, then func.lower_bounds and
%                       func.upper_bounds are set and have the same 
%                       dimensionality as func.cont
%                     - If func.cat is set, then fun.cat_domain_sizes is 
%                       set and has the same dimensionality
%       func.funcHandle: function handle for the blackbox function. Takes 4
%                        arguments: funcHandle(theta, instance_numbers, seeds, censorLimits)
%                        that funcHandle does not actually have to use all
%                        those arguments, see smbo_wrap_simple.m for an
%                        example.
%       
%       func.cont []: N by 1 vector of indices of numerical inputs
%       func.cat []: C by 1 vector of indices of categorical inputs
%
%       func.param_bounds: N+C by 2 vector of lower and upper bounds for numerical inputs
%                           only func.param_bounds(func.cont,:) will be used
%       func.all_values: N+C by 1 vector of domains for categorical inputs
%                           only func.all_values(func.cat) will be used
%                           calls of func.funcHandle will use *indices*
%                           into the domain of each categorical variables,
%                           e.g. one of {1,2,3} for a domain {'a','b','c'} 
%   
%       func.default_values: mandatory input with starting vector
%
%       func.cutoff [inf]:     if not inf, this means the function will accept
%                              a cutoff as input: funcHandle(x, 'cutoff',
%                              cutoff)
%                              in that case, func.cutoff is the maximal
%                              cutoff
%
%       func.features []:    P by d vector of characteristics (features) of 
%                            problem instances ("environmental conditions") 
%                            that we want to optimize a statistic over. 
%                            This is empty if we're just optimizing one 
%                            function (even if it's noisy).
%                            If this is not empty, the function will accept
%                            an additional integer specifying the instance
%                            to use: funcHandle(x, 'instanceIndex',
%                            instanceIndex).
%
%       All additional inputs to funcHandle can be mixed and matched freely.
%
%       func.env []: structure required if func wraps an algorithm run
%
%
%   options: an optional structure containing optional settings
%     Optional fields of options, with default value in square brackets
%       options.seed [1]: seed for a random number generator to ensure repeatability
%       options.maxFunEvals [100*(length(func.cat)+length(func.cont))]
%       options.maxTimeBudget [inf]
%   
% Output parameters: 
%   min_x: 
%     Best found N+C by 1 vector min_x for minimizing funcHandle, where
%          min_x(numerical(n)) in [input_domains.lower_bounds(n),input_domains.upper_bounds(n)]
%          and
%          min_x(categorical(c)) in {1,...,input_domains.cat_domain_sizes(c)}.
%   min_val: value of funcHandle(min_x). Estimation for noisy functions or
%   when dealing with uncontrollable settings to get a statistic over
%   model: for reference, the final model learned

%TODOs: active determination of runlength?
%      error message when combining solqual with mean10
%      capping for solqual: just don't do runs when mean already above lb; should fall out of the code...
% use cached data for validation, even when not using the DB.
if nargin < 1
    fprintf('Usage: smbo(func [, options])\n')
    return
end
if nargin < 2
    options = [];
end
if nargin < 3
    logHeader = '';
end

if ~isfield(options, 'seed')
    options.seed = 1;
end
numRun = options.seed;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP: INITIALIZE RANDOM NUMBER GENERATOR
% Use fixed seed for each run for reproducibility (e.g., there is
% randomness in the local search for optimizing expected improvement).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rand('twister',numRun+1);
randn('state',numRun+1);
s = rand('twister');

func = processFunc(func);
rand('twister', s);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP: GET AC PARAMETERS FROM DEFAULT FILE AND COMMAND LINE CALL.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
options.overallobj = func.overallobj;
options.kappa_max = func.cutoff;
options = set_options(options);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP: INITIALIZE GLOBALS, OUTPUT FILES AND OUTPUT GENERAL HEADER.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
initializeGlobalHouseKeepingVariables();
global writeToScreen;
writeToScreen = options.writeToScreen;
global ThetaUniqSoFar;
[workspace_savedir, paramdir, options, workspace_filename_for_complete_run] = defineOutputFiles(options, func, numRun);
global filename_for_workspace_with_failed_assertion;
filename_for_workspace_with_failed_assertion = parsePath('workspace_with_failed_assertion.mat', workspace_savedir);
global incumbent_matrix; %#ok<NUSED> % declared here, such that it is saved by save
global allParamStrings; %#ok<NUSED> % declared here, such that it is saved by save

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check if a complete run already exists; if so, return that.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if ~isempty(workspace_filename_for_complete_run)
    currentOptions = options;
    currentFunc = func;
    load(workspace_filename_for_complete_run)
    options = currentOptions;
    func = currentFunc;
    return
end
if options.offlineValidationOnly && isempty(workspace_filename_for_complete_run)
    % Only want to do offline validation, but configuration run is not done. Exiting.
    return;
end


if ~isdeployed
    profile off;
    switch options.profile
        case 0
        case 1
            profile on -memory -timer cpu;
        case 2
            profile on -memory -timer real;
        otherwise
            error('unknown profiler type. Options are 0 (off), 1 (cpu), 2(real).');
    end
end

bout(['\n********************************************\nSMBO started at ', datestr(now), '\n********************************************\n'], true);

output_general_header(logHeader, options);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP: GET OFFLINE VALIDATION DATA FOR THE MODEL.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
valdata = getValdata(func, options, numRun);
rand('twister', s);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP: INITIALIZE VARIABLES TO KEEP INTERNAL STATISTICS.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
learnTime1 = [];
iterationTime = [];
ls_time = [];
alTunerTime = 0;
iteration = 0;
options.maxn = min(options.maxn, func.numTrainingInstances*size(func.seeds,2));

max_num_workers = 5000;
options.input_filenames = cell(max_num_workers,1);
options.output_filenames = cell(max_num_workers,1);
for j=1:max_num_workers
    options.input_filenames{j} = ['worker_store/inputs/worker_', num2str(j), '.txt'];
    options.output_filenames{j} = ['worker_store/outputs/worker_', num2str(j), '.txt'];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP: INITIALIZE INSTANCE FEATURES and start workers.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
all_features_for_training_instances = func.features;

worker_ids = options.first_worker_idx:options.first_worker_idx+options.num_sub_workers-1;
% if options.server_can_be_worker
%      worker_ids(end+1) = options.num_sub_workers+1;
% end
% startWorkers(workspace_savedir, worker_ids, options);
%startWorkers(workspace_savedir, [], options);
% worker_ids = 1; % for debugging.
deleteFilesAndReturnWhenFinished(options.input_filenames(worker_ids));
deleteFilesAndReturnWhenFinished(options.output_filenames(worker_ids));
options.workspace_savedir = workspace_savedir;
options.worker_ids = worker_ids;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START WITH INITIAL DATA.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
tout(sprintf('Run %d\n', numRun));
outputNewBest(0, 0, 10000000000000, 0, 1, 0, 0, alphabeticalParameterString(func, func.default_values), 0, 0, 0, options); % 1:default
% [incumbent_theta_idx, bestconflist, rundata, ThetaUniqSoFar] = getInitialData(func, options, numRun);
global bestconflist;
[run_queue, incumbent_theta_idx, bestconflist, rundata, ThetaUniqSoFar, eligible_theta_idxs] = getInitialData(func, options, numRun);

[run_queue, used_worker_ids, assigned_runs, output_filenames] = assignRunsToRubyWorkers(run_queue, options, func, worker_ids);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fill up run_queue for next iteration with random runs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
numChallengersNeeded = computeNumChallengersNeeded(options, run_queue, incumbent_theta_idx);
while numChallengersNeeded > 0
    %== Select challengers at random.
    challengerStore = selectRandomConfigs(func, numChallengersNeeded);
    
    tmp_rundata = hallucinate_run_queue_finished(rundata, run_queue, func);
    [run_queue, eligible_theta_idxs] = determineNewRuns(0, challengerStore, incumbent_theta_idx, func, tmp_rundata, options, run_queue, eligible_theta_idxs);
    numChallengersNeeded = computeNumChallengersNeeded(options, run_queue, incumbent_theta_idx);
end

rand('twister', s); % such that any prior randomness does not change the seed
%try

iteration=0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ITERATE UNTIL OUT OF TIME.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
while ~have_to_stop(alTunerTime, rundata.runTime, func, rundata.used_instance_idxs, options, iteration)
    %=== Set up some housekeeping variables.
    iteration = iteration+1;
    valdata.iteration = iteration;
    iterationTime(iteration) = 0;
    figure_prefix = strcat(paramdir, '/run', num2str(numRun), '/it', num2str(iteration), '-');
    valdata.figure_prefix = figure_prefix;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Waiting for worker nodes to finish.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    bout(['Iteration ', num2str(iteration), ': waiting for runs to finish...\n']);
    tic;
    [finished_runs_with_results, output_filenames_to_delete, used_worker_ids, assigned_runs, output_filenames] = waitForRubyWorkers(iteration, used_worker_ids, assigned_runs, output_filenames, func);
    waitingTime = toc
    tic;
    deleteFilesAndReturnWhenFinished(output_filenames_to_delete)
    deleteTime = toc

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Process finished runs.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [rundata, run_queue, uniq_new_theta_idxs] = update_rundata_with_new_runs(rundata, run_queue, finished_runs_with_results, iteration);
    if iterationTime(end) > rundata.algoWallTimePerIteration(end)
        bout(['WARNING: model learning and EI optimization took longer (', num2str(iterationTime(end)), ') than the parallel algorithm runs (', num2str(rundata.algoWallTimePerIteration(end)),'). Counting that longer time instead.']);
        rundata.algoWallTimePerIteration(end) = iterationTime(end);
    end
    my_assert(size(run_queue(:,[1,2,3]),1)==size(unique(run_queue(:,[1,2,3]),'rows'),1), 'Asserting that we do not create duplicate runs');

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Update incumbent (based on finished runs).
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [incumbent_theta_idx, eligible_theta_idxs] = update_incumbent(rundata, incumbent_theta_idx, eligible_theta_idxs, run_queue, options, func, alTunerTime, iteration, 0);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % For challengers with no runs left in the queue, decide whether to 
    % queue additional runs (and if so whether they are eligible to become 
    % the incumbent once those runs are done).
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    for i=length(uniq_new_theta_idxs):-1:1 % iterate over this list backwards to get runs in right order when putting them at start of queue
        if isempty(intersect(find(run_queue(:,1)==uniq_new_theta_idxs(i)), find(run_queue(:,5) < 2))) 
            [run_queue, eligible_theta_idxs] = determineNewRuns(1, ThetaUniqSoFar(uniq_new_theta_idxs(i),:), incumbent_theta_idx, func, rundata, options, run_queue, eligible_theta_idxs);
            my_assert(size(run_queue(:,[1,2,3]),1)==size(unique(run_queue(:,[1,2,3]),'rows'),1), 'Asserting that we do not create duplicate runs');
        end
    end
    bout(['Configurations determined eligible at iteration ', num2str(iteration), ': [']);
    for i=1:length(eligible_theta_idxs)
        bout([num2str(eligible_theta_idxs(i)), ' ']);
    end            
    bout(']\n');

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Assign prepared runs to the worker nodes.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [run_queue, new_used_worker_ids, new_assigned_runs, new_output_filenames] = assignRunsToRubyWorkers(run_queue, options, func, setdiff(worker_ids, used_worker_ids));
    used_worker_ids = [used_worker_ids; new_used_worker_ids];
    assigned_runs = [assigned_runs; new_assigned_runs];
    output_filenames = [output_filenames; new_output_filenames];
    my_assert(size(run_queue(:,[1,2,3]),1)==size(unique(run_queue(:,[1,2,3]),'rows'),1), 'Asserting that we do not create duplicate runs');

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Get challenger runs to fill up the next iteration.
    % In some borderline cases, the run_queue might not be entirely
    % full after the first filling-up
    % (e.g. when selecting incumbent and not doing bonus run, or when 
    % selecting another configruation that already has maxn runs).
    % In those cases, fill up with random challengers.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    numChallengersNeeded = computeNumChallengersNeeded(options, run_queue, incumbent_theta_idx);
    selectChallengerIteration = 1;
    model.constant = 1; % make a first model to pass
    model.lastTimeToLearn = 0;
    while numChallengersNeeded > 0
        %=== Select challengers.
        if selectChallengerIteration == 1 && strcmp(options.method, 'al')
            %== Use the model to select challengers.
            [model, valdata, valStats, learnTime1, alTunerTime, iterationTime] = getModel(model, rundata, all_features_for_training_instances, options, func, iteration, numRun, valdata, figure_prefix, alTunerTime, iterationTime);
            [means, vars] = applyMarginalModel(model, ThetaUniqSoFar, 1:length(model.module), [], 0, 0);
            [challengerStore, eiTime, valStats] = select_all_challengers(func, model, means, vars, options, incumbent_theta_idx, valdata, valStats, rundata, learnTime1, numChallengersNeeded);

            ls_time(iteration) = eiTime;
            if (options.timeout_incl_learning)
                alTunerTime = alTunerTime + ls_time(iteration);
            end
            iterationTime(iteration) = iterationTime(iteration) + ls_time(iteration);
        else
            %== Select challengers at random.
            challengerStore = selectRandomConfigs(func, numChallengersNeeded);
        end

        %=== Fill up run_queue with challengerStore and the new incumbent bonus runs for considering those new configs.
        if ~isempty(challengerStore)
            tmp_rundata = hallucinate_run_queue_finished(rundata, run_queue, func);
            [run_queue, eligible_theta_idxs] = determineNewRuns(0, challengerStore, incumbent_theta_idx, func, tmp_rundata, options, run_queue, eligible_theta_idxs);
            
            tmp_rundata = hallucinate_run_queue_finished(rundata, run_queue, func);
            run_queue = do_additional_incumbent_runs(tmp_rundata, run_queue, incumbent_theta_idx, options, size(challengerStore,1), func);
            my_assert(size(run_queue(:,[1,2,3]),1)==size(unique(run_queue(:,[1,2,3]),'rows'),1), 'Asserting that we do not create duplicate runs');
        end
        selectChallengerIteration = selectChallengerIteration+1;
        numChallengersNeeded = computeNumChallengersNeeded(options, run_queue, incumbent_theta_idx);
    end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Wrap up.    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%catch ME
%     endWorkers(workspace_savedir, worker_ids);
%    rethrow(ME);
%end
valdata.next_iteration_to_output = iteration;
outputCurrentStatisticsAndRunFiles(workspace_savedir, ThetaUniqSoFar, rundata, valdata, iteration, numRun, options);

%== Get empirical mean and variance of the mean to output.
data = rundata.y(rundata.used_theta_idxs==incumbent_theta_idx)';
min_val = combineRunObjectives(options.overallobj, data, func.cutoff);
min_x = config_back_transform(ThetaUniqSoFar(incumbent_theta_idx,:), func);

bout(['\n********************************************\nSMBO finished at ', datestr(now), '. Final best objective: ', num2str(min_val), '\nBest configuration:\n'], true);
paramString = alphabeticalParameterString(func, min_x, true);
bout([paramString, '\n********************************************\n'], true);

workspace_filename = [options.workspace_filenameprefix, '_end_of_SMBO', '.mat'];
bout(sprintf(['Saving workspace at end of SMBO run to ' workspace_filename '\n']));
save(workspace_filename);

function numChallengersNeeded = computeNumChallengersNeeded(options, run_queue, incumbent_theta_idx)
% Computes how many new challengers we need: in the worst case,
% if in the next iteration we don't want to do any bonus runs,
% we still need to fill up the workers next iteration.
%
% We'll typically do one new run for the incumbent for every new challenger, but only up to a point.
numRunsUnstarted = length(find(run_queue(:,5)==0));
numRunsNeeded = options.num_sub_workers*options.num_episode - numRunsUnstarted;
maxNumNewBonusRunsIncumbent = options.maxn - length(find(run_queue(:,1)==incumbent_theta_idx));
numChallengersNeeded = max(ceil(numRunsNeeded/2), numRunsNeeded-maxNumNewBonusRunsIncumbent); 

function rundata = hallucinate_run_queue_finished(rundata, run_queue, func)
%=== Fill rundata up with incumbent runs that are not run yet, to avoid duplicates.
idxs = find(run_queue(:,5)<2);
switch func.singleRunObjective
    case 'runtime'
        hallucinated_value = func.cutoff+1e-6;
    case 'runlength'    
        hallucinated_value = 1e10+1;
    case 'solqual'
        hallucinated_value = 1e10+1;
    otherwise 
        error 'Still have to implement objective functions for single algorithm runs other than runtime, runlength, and solqual.'
end
    
for i=1:length(idxs)    
    rundata = hallucinate_rundata_with_new_y(hallucinated_value, rundata, run_queue(idxs(i),1), run_queue(idxs(i),2), run_queue(idxs(i),4));
end